function init(dataid,w,h)
	gdata = mmf.newObject(dataid)
	--[[
	map = {}
	movestack = {}
	]]--
	undostack = {}
	screenw = w
	screenh = h
	units = {}
	
	loadlevel(1)
end

ID = 1
XPOS = 4
YPOS = 5
TYPE = 6
POSTYPE = 7
SPRITE = 8

POSCORRECT = 23

SIZE = 1
XMAP = 2
YMAP = 3
MIDX = 4
MIDY = 5
LEVEL = 6
SHAKING = 7
WINTIMER = 8

ID_CURR = 0

funcs =
{
	knight = function(unit,ox,oy,dir)
		local x,y = unit.values[XPOS], unit.values[YPOS]
		local opts = {{{2,1},{2,-1}},{{1,-2},{-1,-2}},{{-2,-1},{-2,1}},{{-1,2},{1,2}}}
		local opt = opts[dir+1]
		local nx,ny = x + ox,y + oy
		local furthest = -1
		
		if (opt ~= nil) then
			local tx,ty = nx,ny
			local this = 0
			
			for i,o in ipairs(opt) do
				local ix = x + o[1]
				local iy = y + o[2]
				local dist = finddist(ix,iy)
				local target = findunit(ix,iy)
				
				if (dist > furthest) then
					if (target == nil) then
						furthest = dist
						this = i
					elseif (target.values[TYPE] == 2) then
						local p,pp = trypush(target,{},ox,oy,dir,1)
						
						if pp then
							furthest = dist
							this = i
						end
					end
				end
			end
			
			if (this > 0) then
				local o = opt[this]
				local tx = x + o[1]
				local ty = y + o[2]
				
				return tx,ty
			end
		end
		
		return nx,ny
	end,
	pawn = function(unit,ox,oy,dir)
		local x,y = unit.values[XPOS], unit.values[YPOS]
		local opts = {{{1,1},{1,-1}},{{1,-1},{-1,-1}},{{-1,-1},{-1,1}},{{-1,1},{1,1}}}
		local opt = opts[dir+1]
		local nx,ny = x + ox,y + oy
		local furthest = -1
		
		if (opt ~= nil) then
			local tx,ty = nx,ny
			local this = 0
			
			for i,o in ipairs(opt) do
				local ix = x + o[1]
				local iy = y + o[2]
				local dist = finddist(ix,iy)
				local target = findunit(ix,iy)
				
				--MF_alert(tostring(ix) .. ", " .. tostring(iy) .. ", " .. tostring(dist))
				
				if (dist > furthest) then
					if (target == nil) then
						furthest = dist
						this = i
					elseif (target.values[TYPE] == 2) then
						local p,pp = trypush(target,{},ox,oy,dir,1)
						
						if pp then
							furthest = dist
							this = i
						end
					end
				end
			end
			
			if (this > 0) then
				local o = opt[this]
				local tx = x + o[1]
				local ty = y + o[2]
				
				return tx,ty
			end
		end
		
		return nx,ny
	end,
}

function createunit(name,x,y)
	local d = things[name] or {}
	create("Object",x,y,d)
end

things =
{
	player = 
	{
		sprite = 0,
		player = true,
		icon = "p",
	},
	target = 
	{
		sprite = 5,
		pushtype = -1,
		icon = "T",
	},
	wall = 
	{
		sprite = 1,
		pushtype = 0,
		icon = "W",
	},
	box = 
	{
		sprite = 2,
		pushtype = 1,
		icon = "B",
	},
	knight = 
	{
		sprite = 3,
		pushtype = 1,
		push = funcs.knight,
		icon = "K",
	},
	pawn = 
	{
		sprite = 4,
		pushtype = 1,
		push = funcs.pawn,
		icon = "P",
	},
	tile = 
	{
		sprite = 6,
		pushtype = -2,
		icon = "!",
	},
	box_multi = 
	{
		sprite = 7,
		pushtype = 2,
		icon = "b",
	},
	knight_multi = 
	{
		sprite = 8,
		pushtype = 2,
		push = funcs.knight,
		icon = "k",
	},
	pawn_multi = 
	{
		sprite = 9,
		pushtype = 2,
		push = funcs.pawn,
		icon = "q",
	},
}

levels =
{
	{
		title = "RUDE CHESS",
		subtitle = "Fill every X with a piece; when near a piece, the red square indicates where it would move if pushed.\r\nUse Arrows to move, Z to undo, R to restart, ctrl + D to skip level.",
		ps = "Made by Arvi Teikari at NGJ 2022 (Version 1.2b)",
		"WWWWWWWWW",
		"W.......W",
		"W..W....W",
		"W.....T.W",
		"W..W..T.W",
		"W..W..p.W",
		"W...B.W.W",
		"W..WBWW.W",
		"W.......W",
		"WWWWWWWWW",
	},
	{
		title = "PAWNS",
		subtitle = "Pawns move diagonally if possible.\r\nIf not possible, they move forward like boxes.",
		"WWW....",
		"WpW....",
		"WPW....",
		"W.W....",
		"W.WW...",
		"W..WW..",
		"W.WTW..",
		"WPWWWWW",
		"W....TW",
		"WWWWWWW",
	},
	{
		title = "HOW RUDE",
		subtitle = "When given a choice, pieces prefer moving away from the Xs.",
		".WWWWWW.",
		".W....W.",
		".WPW..W.",
		".W.W..W.",
		".W....WW",
		".W.WB.TW",
		".W.p..WW",
		"WW.W..W.",
		"WTWW..W.",
		"WWWWWWW.",
	},
	{
		".WWWWWWWWW.",
		".W.......W.",
		".W.......W.",
		".W...P...W.",
		"WW...B...WW",
		"W..T.p.T..W",
		"WW.......WW",
		".W.......W.",
		".WWWWWWWWW.",
	},
	{
		title = "KNIGHTS",
		subtitle = "Knights do leaps and prefer moving away from the Xs.\r\nIf a knight can't leap, it'll move forward like a box.",
		"WWWWWWW",
		"W..K..W",
		"W.W...W",
		"W.W.T.W",
		"W.W...W",
		"WWW.WWW",
		"W.....W",
		"W..p..W",
		"W.....W",
		"WWWWWWW",
	},
	{
		"WWWWWWW",
		"W.....W",
		"W.pBK.W",
		"W.....W",
		"W..T..W",
		"W.....W",
		"WW.T..W",
		"W.....W",
		"WW.W.WW",
		"WWWWWWW",
	},
	{
		".WWWWWWW",
		".W.....W",
		".W.K.p.W",
		".W.....W",
		".W.....W",
		"WW.....W",
		"W..T...W",
		"WW.....W",
		".WWWWWWW",
	},
	{
		"WWWWWWWWW",
		"W.......W",
		"W..W....W",
		"W..p..T.W",
		"W..W..T.W",
		"W..W.K..W",
		"W..B..W.W",
		"W..W.WW.W",
		"W.......W",
		"WWWWWWWWW",
	},
	{
		".WWWWWWWWW",
		".W.......W",
		".W.......W",
		".W...P...W",
		"WW...KWWWW",
		"W..T.pWTWW",
		"WW....WWWW",
		".W.......W",
		".WWWWWWWWW",
	},
	{
		title = "MULTIPUSH?",
		subtitle = "Pieces marked with a plus sign can be pushed together with other pieces.",
		"WWW......",
		"WpW......",
		"W.WWWW...",
		"WBWWqWW..",
		"WbWq.qWWW",
		"Wbq.W.qTW",
		"W.WBWWWWW",
		"W.W.W....",
		"W.WTW....",
		"WWWWW....",
	},
	{
		"WWWWWWT",
		"W...W.W",
		"W....WW",
		"W.qB..W",
		"W.pq..W",
		"W.....W",
		"WWWWWWW",
	},
	{
		"WWWWWWW",
		"W....TW",
		"WB....W",
		"W..bWWW",
		"W..W.TW",
		"Wq.W.WW",
		"W...p.W",
		"WWW...W",
		"..WWWWW",
	},
	{
		"WTWWWWW",
		"TW....W",
		"W.....W",
		"W.....W",
		"W.K.q.W",
		"W.p...W",
		"WWWWWWW",
	},
	{
		"WWWWWWT",
		"W.B..TW",
		"W....TW",
		"W.k..WW",
		"W.pk..W",
		"W.....W",
		"WWWWWWW",
	},
	{
		title = "You win!",
		subtitle = "Thank you for playing!",
		ps = "Made by Arvi Teikari at NGJ 2022 using Multimedia Fusion 2",
		".........",
		".KkPqPkK.",
		".BbBbBbB.",
		".........",
		"....p....",
		".........",
		".BbBbBbB.",
		".KkPqPkK.",
		".........",
	},
}

function loadlevel(lvl)
	local lvlid = lvl or gdata.values[LEVEL]
	
	local d = levels[lvlid]
	
	if (d ~= nil) then
		gdata.values[XMAP] = #d[1]
		gdata.values[YMAP] = #d
		
		gdata.values[MIDX] = screenw * 0.5 - gdata.values[XMAP] * gdata.values[SIZE] * 0.5
		gdata.values[MIDY] = screenh * 0.5 - gdata.values[YMAP] * gdata.values[SIZE] * 0.5
		
		for j,v in ipairs(d) do
			for i=1,#v do
				local l = string.sub(v, i, i)
				
				for a,b in pairs(things) do
					if (b.icon == l) then
						createunit(a,i-1,j-1)
					end
				end
				
				createunit("tile",i-1,j-1)
			end
		end
	else
		d = {}
	end
	
	local t1 = d.title or ""
	local t2 = d.subtitle or ""
	local t3 = d.ps or ""
	
	MF_title(t1,t2,t3)
	
	dotargets(true)
end

function dotargets(instant_)
	local instant = instant_ or false
	
	for i,unit in ipairs(units) do
		if (unit.player ~= nil) and unit.player then
			local dirs = {{1,0},{0,-1},{-1,0},{0,1}}
			for d=0,3 do
				local drs = dirs[d+1]
				local dx,dy = drs[1],drs[2]
				move(unit,dx,dy,d,true,instant)
			end
		end
	end
end

function newid()
	ID_CURR = ID_CURR + 1
	return ID_CURR
end

function create(name,x,y,data)
	local unitid = MF_create(name)
	local unit = addunit(unitid)
	unit.values[XPOS] = x
	unit.values[YPOS] = y
	unit.values[POSCORRECT] = 8
	unit.values[ID] = newid()
	
	if (data ~= nil) then
		for i,v in pairs(data) do
			unit[i] = v
		end
	end
	
	unit.values[TYPE] = unit.pushtype or 0
	unit.values[POSTYPE] = unit.postype or 0
	unit.values[SPRITE] = unit.sprite or 0
end

function addunit(unitid)
	units[#units+1] = mmf.newObject(unitid)
	local unit = units[#units]
	return unit
end

function findunit(x,y)
	for i,unit in ipairs(units) do
		if on(unit,x,y) and (unit.values[TYPE] >= 0) and (unit.player == nil) then
			return unit
		end
	end
	
	return nil
end

function findtarget(x,y)
	for i,unit in ipairs(units) do
		if on(unit,x,y) and (unit.values[TYPE] == -1) then
			return unit
		end
	end
	
	return nil
end

function finddist(x,y)
	local dist = 999
	for i,unit in ipairs(units) do
		if (unit.values[TYPE] == -1) then
			local ux,uy = unit.values[XPOS], unit.values[YPOS]
			local udist = math.sqrt((uy - y) ^ 2 + (ux - x) ^ 2)
			
			if (math.abs(x - ux) > 1) or (math.abs(y - uy) > 1) then
				udist = udist + 5
			end
			
			dist = dist + udist
		end
	end
	
	return dist
end

function on(unit,x,y)
	return (unit.values[XPOS] == x) and (unit.values[YPOS] == y)
end

function move(unit,ox,oy,dir,sim_,instant_)
	local x,y = unit.values[XPOS], unit.values[YPOS]
	local nx,ny = x + ox, y + oy
	
	local push = {}
	local moving = false
	local target = findunit(nx,ny)
	local sim = sim_ or false
	local instant = instant_ or false
	
	if (target == nil) then
		push = {{unit, nx, ny}}
		moving = true
	elseif (target.values[TYPE] >= 1) then
		local pushing = false
		push = {{unit, nx, ny}}
		push,pushing = trypush(target,push,ox,oy,dir,1)
		
		if pushing then
			moving = true
		end
	end
	
	if moving then
		for i,v_ in ipairs(push) do
			local v = v_[1]
			
			if (sim == false) then
				addundo({"move",v.values[ID],v.values[XPOS],v.values[YPOS]})
				v.values[XPOS] = v_[2]
				v.values[YPOS] = v_[3]
				v.values[POSCORRECT] = 0
			else
				if (v.fixed ~= unit.fixed) then
					local unitid = MF_create("Target")
					local u = mmf.newObject(unitid)
					u.x = v.x
					u.y = v.y
					u.values[XPOS] = v_[2]
					u.values[YPOS] = v_[3]
					
					if instant then
						u.values[POSCORRECT] = 8
					end
				end
			end
		end
	end
end

function trypush(unit,push_,ox,oy,dir,pushid)
	local x,y = unit.values[XPOS], unit.values[YPOS]
	local nx,ny = x + ox, y + oy
	local push = {}
	
	if (unit.push ~= nil) then
		nx,ny = unit.push(unit,ox,oy,dir)
	end
	
	for i,v in ipairs(push_) do
		table.insert(push, {v[1], v[2], v[3]})
	end
	
	local moving = false
	local target = findunit(nx,ny)
	
	if (target == nil) then
		moving = true
	elseif (target.values[TYPE] >= 1) then
		local pushing = false
		
		push,pushing = trypush(target,push,ox,oy,dir,pushid + 1)
		
		if pushing and ((unit.values[TYPE] ~= 1) or ((unit.values[TYPE] == 1) and (pushid <= 1))) then
			moving = true
		end
	end
	
	if ((unit.values[TYPE] == 1) and (pushid > 1)) then
		moving = false
	end
	
	if moving then
		table.insert(push, {unit, nx, ny})
	end
	
	return push,moving
end

function command(ox,oy,dir)
	MF_cleartargets()
	newundo()
	
	for i,v in ipairs(units) do
		if (v.player ~= nil) and v.player then
			move(v,ox,oy,dir)
			
			local dirs = {{1,0},{0,-1},{-1,0},{0,1}}
			for d=0,3 do
				local drs = dirs[d+1]
				local dx,dy = drs[1],drs[2]
				move(v,dx,dy,d,true)
			end
		end
	end
	
	wincheck()
end

function newundo()
	if (#undostack > 0) and (#undostack[1] == 0) then
		table.remove(undostack, 1)
	end
	
	table.insert(undostack, 1, {})
end

function addundo(data)
	local cundo = undostack[1]
	table.insert(cundo, {})
	local this = cundo[#cundo]
	
	for i,v in ipairs(data) do
		table.insert(this, v)
	end
end

function undo()
	MF_cleartargets()
	
	if (#undostack > 0) then
		local cundo = undostack[1]
		
		for i,v in ipairs(cundo) do
			if (v[1] == "move") then
				for a,b in ipairs(units) do
					if (b.values[ID] == v[2]) then
						b.values[XPOS] = v[3]
						b.values[YPOS] = v[4]
						b.values[POSCORRECT] = 0
					end
				end
			end
		end
		
		table.remove(undostack, 1)
	end
	
	dotargets()
end

function restart()
	clear()
	loadlevel()
end

function wincheck()
	local win = true
	local valid = false
	
	for i,unit in ipairs(units) do
		if (unit.values[TYPE] == -1) then
			valid = true
			
			local trg = findunit(unit.values[XPOS], unit.values[YPOS])
			if (trg == nil) then
				win = false
			end
		end
	end
	
	if win and valid then
		dowin()
	end
end

function dowin()
	MF_win()
end

function nextlevel()
	clear()
	gdata.values[LEVEL] = gdata.values[LEVEL] + 1
	loadlevel()
end

function clear()
	units = {}
	undostack = {}
	ID_CURR = 0
	MF_clear()
end